# -*- coding: utf-8 -*-

import asyncio
import pytest
import pytest_asyncio

from patchwork.core import Task
from patchwork.core.client.local import AsyncLocalPublisher, AsyncLocalSubscriber, AsyncLocalBroker, MissingBehaviour
from patchwork.core.client.serializers.betterproto import BetterprotoSerializer


@pytest.fixture
def broker():
    return AsyncLocalBroker(
        ('test', ),
        publish_missing=MissingBehaviour.EXCEPTION,
        subscribe_missing=MissingBehaviour.EXCEPTION
    )


@pytest_asyncio.fixture
async def subscriber(broker):
    client = AsyncLocalSubscriber(broker=broker, queue_names=['test'])
    async with client:
        yield client


@pytest.mark.asyncio
async def test_send(broker):
    client = AsyncLocalPublisher(broker=broker)

    await client.send(b'payload', some_extra=1, queue_name='test')

    assert broker.get_queue('test').qsize() == 1, \
        "Sent message missing on the queue"

    payload = broker.get_queue('test').get_nowait()
    serializer = BetterprotoSerializer()
    task = serializer.decode(payload)

    assert task.payload == b'payload'
    assert task.meta.extra['some_extra'] == 1


@pytest.mark.asyncio
async def test_send_payload(broker):
    client = AsyncLocalPublisher(broker=broker)

    task = Task(task_type='test')
    task.meta.queue_name = 'test'
    await client._send(b'payload', task)

    assert broker.get_queue('test').qsize() == 1, \
        "Sent message missing on the queue"

    payload = broker.get_queue('test').get_nowait()
    assert payload == b'payload'


@pytest.mark.asyncio
async def test_send_timeout():
    broker = AsyncLocalBroker(['test'], max_queue_size=1)
    # put something to fill the queue
    broker.get_queue('test').put_nowait(None)

    client = AsyncLocalPublisher(broker=broker)

    with pytest.raises(TimeoutError):
        task = Task(task_type='test')
        task.meta.queue_name = 'test'
        await client._send(b'payload', task, timeout=0.01)


@pytest.mark.asyncio
async def test_shared_default_queue():
    """
    Test if all local async client instances share the same default broker instance
    :return:
    """
    client_1 = AsyncLocalPublisher()
    client_2 = AsyncLocalSubscriber(queue_names=['test'])

    task = Task(task_type='test')
    task.meta.queue_name = 'test'
    await client_1._send(b'payload', task)
    async with client_2:
        payload, meta = await client_2._fetch_one()

    assert payload == b'payload'
    assert meta == {'queue_name': 'test'}


@pytest.mark.asyncio
async def test_fetch_one(subscriber, broker):
    broker.get_queue('test').put_nowait(b'payload')

    payload, meta = await subscriber._fetch_one()

    assert payload == b'payload'
    assert meta == {'queue_name': 'test'}


@pytest.mark.asyncio
async def test_fetch_one_timeout(subscriber):
    with pytest.raises(TimeoutError):
        _payload, _meta = await subscriber._fetch_one(timeout=0.01)


@pytest.mark.asyncio
async def test_unstarted_subscriber(broker):
    client = AsyncLocalSubscriber(broker=broker, queue_names=['test'])

    with pytest.raises(RuntimeError):
        _ = await client._fetch_one()


@pytest.mark.asyncio
async def test_subscribe_on_missing():
    """
    Test if subscriber subscribed on missing queue will receive a message when queue will be created
    :return:
    """
    broker = AsyncLocalBroker()
    client_1 = AsyncLocalPublisher(broker=broker)
    client_2 = AsyncLocalSubscriber(broker=broker, queue_names=['test-missing'])

    async def send():
        task = Task(task_type='test')
        task.meta.queue_name = 'test-missing'
        await client_1._send(b'payload', task)

    # create a background task to send task on missing queue (will be created)
    asyncio.create_task(send())

    async with client_2:
        # await on subscribed queue which does not exists yet
        assert not client_2.broker.has_queue('test-missing')
        payload, meta = await client_2._fetch_one()

    assert payload == b'payload'
    assert meta == {'queue_name': 'test-missing'}



# -*- coding: utf-8 -*-
# Note: this file is a port of pending PR on Pydantic (https://github.com/samuelcolvin/pydantic/pull/3159)
# about adding support for nested settings in ENV settings source
# https://github.com/samuelcolvin/pydantic/blob/1994a926bcdd14aac008dac00dcd779a37a61880/pydantic/env_settings.py

import sys

from pathlib import Path

import os

from pydantic import BaseSettings
from pydantic.typing import get_origin
from pydantic.env_settings import read_env_file, SettingsError
from typing import Optional, Mapping, Dict, Any, Type, Union


try:
    from typing import GenericAlias as TypingGenericAlias  # type: ignore
except ImportError:
    # python < 3.9 does not have GenericAlias (list[int], tuple[str, ...] and so on)
    TypingGenericAlias = ()


StrPath = Union[str, os.PathLike]


if sys.version_info < (3, 10):

    def is_union_origin(tp: Type[Any]) -> bool:
        return tp is Union

    WithArgsTypes = (TypingGenericAlias,)

else:
    import types
    import typing

    def is_union_origin(origin: Type[Any]) -> bool:
        return origin is Union or origin is types.UnionType  # noqa: E721

    WithArgsTypes = (typing._GenericAlias, types.GenericAlias, types.UnionType)


class NestedEnvSettingsSource:
    __slots__ = ('env_file', 'env_file_encoding', 'env_nested_delimiter')

    def __init__(
        self, env_file: Optional[StrPath], env_file_encoding: Optional[str], env_nested_delimiter: Optional[str]
    ):
        self.env_file: Optional[StrPath] = env_file
        self.env_file_encoding: Optional[str] = env_file_encoding
        self.env_nested_delimiter: Optional[str] = env_nested_delimiter

    def get_env_vars(self, settings: BaseSettings) -> Mapping[str, Optional[str]]:
        if settings.__config__.case_sensitive:
            env_vars: Mapping[str, Optional[str]] = os.environ
        else:
            env_vars = {k.lower(): v for k, v in os.environ.items()}

        if self.env_file is None:
            self.env_file = settings.__config__.env_file

        if self.env_file_encoding is None:
            self.env_file_encoding = settings.__config__.env_file_encoding

        if self.env_nested_delimiter is None:
            self.env_nested_delimiter = settings.__config__.env_nested_delimiter or '__'

        if self.env_file is not None:
            env_path = Path(self.env_file).expanduser()
            if env_path.is_file():
                env_vars = {
                    **read_env_file(
                        env_path, encoding=self.env_file_encoding, case_sensitive=settings.__config__.case_sensitive
                    ),
                    **env_vars,
                }
        return env_vars

    def explode_env_vars(self, settings: BaseSettings, env_vars: Mapping[str, Optional[str]]) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        for env_name, env_val in env_vars.items():
            keys = env_name.split(self.env_nested_delimiter)
            env_var = result
            for idx, key in enumerate(keys):
                if idx == len(keys) - 1:
                    try:
                        env_val = settings.__config__.json_loads(env_val)  # type: ignore
                    except (ValueError, TypeError):
                        ...
                    env_var[key] = env_val
                else:
                    env_var = env_var.setdefault(key, {})
        return result

    def __call__(self, settings: BaseSettings) -> Dict[str, Any]:
        """
        Build environment variables suitable for passing to the Model.
        """
        d: Dict[str, Optional[str]] = {}

        env_vars = self.get_env_vars(settings)
        if self.env_nested_delimiter is not None:
            env_vars = self.explode_env_vars(settings, env_vars)

        for field in settings.__fields__.values():
            env_val: Optional[str] = None
            for env_name in field.field_info.extra['env_names']:
                env_val = env_vars.get(env_name)
                if env_val is not None:
                    break

            if env_val is None:
                continue

            if field.is_complex():
                if isinstance(env_val, (str, bytes, bytearray)):
                    try:
                        env_val = settings.__config__.json_loads(env_val)
                    except ValueError as e:
                        raise SettingsError(f'error parsing JSON for "{env_name}"') from e
            elif (
                is_union_origin(get_origin(field.type_))
                and field.sub_fields
                and any(f.is_complex() for f in field.sub_fields)
            ):
                if isinstance(env_val, (str, bytes, bytearray)):
                    try:
                        env_val = settings.__config__.json_loads(env_val)
                    except ValueError:
                        pass
            d[field.alias] = env_val
        return d

    def __repr__(self) -> str:
        return f'NestedEnvSettingsSource(env_file={self.env_file!r}, env_file_encoding={self.env_file_encoding!r})'


nested_env_settings = NestedEnvSettingsSource(None, None, None)



# -*- coding: utf-8 -*-

import asyncio
import warnings
import weakref
from typing import Literal, Callable, Dict, List

import logging
from pydantic import ValidationError, BaseModel

from patchwork.core.utils import cached_property, Flag


class OrphanedComponent(RuntimeError):
    """
    Component which has parent defined becomes orphaned which means that parent has been destroyed.
    """
    pass


class Component:
    """
    A base class for every Patchwork component.
    """

    class Config(BaseModel):
        """
        Base settings class for all components

        :cvar enabled: determines if component is enabled or not
        """
        enabled: bool = True

    EVENTS = ('startup', 'shutdown')

    settings: Config
    logger_name: str
    state: Flag

    def __init__(self, *, parent=None, **options):
        """
        :param parent:  optional parent component
        :param options: component configuration which must conform `Settings` schema
        """
        if parent is not None:
            self._parent = weakref.ref(parent)

        try:
            self.settings = self.Config(**options)
        except ValidationError as e:
            from patchwork.core.config.base import ImproperlyConfigured
            raise ImproperlyConfigured(f"{self.__class__.__name__}: {e}\nsettings received: {options}") from e

        self.state = Flag(False)

        self._is_stopping: bool = False
        self._is_starting: bool = False

        self._event_handlers: Dict[str, List[Callable]] = {
            event_name: [] for event_name in self.EVENTS
        }

    def __repr__(self):
        state = 'running' if self.is_running else ('starting' if self.is_starting else (
            'stopping' if self.is_stopping else 'stopped'))
        return f"<Component {self.__class__.__name__}[{hex(id(self))}] {state}>"

    async def __aenter__(self):
        await self.run()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.terminate()

    @property
    def parent(self):
        """
        Returns parent instance
        :return:
        """
        if self._parent is None:
            return None

        parent = self._parent()
        if parent is None:
            raise OrphanedComponent("Parent instance gone! This component instance is orphaned and should be destroyed")

        return parent

    @property
    def is_running(self):
        """
        Tells if component is running
        :return: True if component is running, False otherwise
        """
        if not hasattr(self, 'state'):
            return False
        return self.state.value

    @property
    def is_starting(self):
        """
        Tells if component is starting
        :return: True if component is in the middle of starting process, False otherwise
        """
        return self._is_starting

    @property
    def is_stopping(self):
        """
        Tells if component is stopping
        :return: True if component is in the middle of stopping process, False otherwise
        """
        return self._is_stopping

    @cached_property
    def logger(self) -> logging.Logger:
        """
        Returns logger instance for the component
        :return:
        """
        return logging.getLogger(f'patchwork.{getattr(self, "logger_name", self.__class__.__name__.lower())}')

    def _add_event_handler(self, event: str, handler: Callable):
        if event not in self.EVENTS:
            raise ValueError(f'unknown event {event}')
        self._event_handlers[event].append(handler)

    async def _dispatch_event(self, event: str, **kwargs):
        assert event in self.EVENTS, "no such event to dispatch"
        for handler in self._event_handlers[event]:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(**kwargs)
                else:
                    handler(**kwargs)
            except Exception as e:
                warnings.warn(f'event handler failed with exception: {e.__class__.__name__}({e})', RuntimeWarning)

    def on_event(self, event: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self._add_event_handler(event, func)
            return func

        return decorator

    async def _start(self) -> bool:
        """
        Starts component internals. Implement this method in your component.

        !!! hint
            This method internally should handle all known or expected exceptions, make possible cleanup and
            return False. Raise exceptions from here only if error in unrecoverable (aka fatal)

        :return: True if startup succeed, False otherwise
        """
        # put custom startup code here, return from this method when all related objects are started and running
        # await if needed!
        pass

    async def run(self) -> bool:
        """
        Starts component

        !!! note
            If component is already started this method returns `False`!

        :return: True if has been started, False otherwise
        """
        if self.is_running:
            return False

        if not self.settings.enabled:
            self.logger.error("Can't start disabled component")
            raise RuntimeError("Module disabled")

        self._is_starting = True

        if await self._start() is False:
            self._is_starting = False
            return False

        await self._dispatch_event('startup')

        self.state.set(True)
        self._is_starting = False

        return True

    async def terminate(self) -> bool:
        """
        Stops component

        !!! note
            If the component is already stopped this method returns `False`!

        :return: True if has been stopped, False otherwise
        """
        if not self.is_running:
            return False

        self._is_stopping = True

        await self._dispatch_event('shutdown')

        if await self._stop() is False:
            self._is_stopping = False
            return False

        self.state.set(False)
        self._is_stopping = False
        return True

    async def _stop(self) -> bool:
        """
        Stops component internals. Implement this method in your component.

        !!! hint
            This method internally should handle all known or expected exceptions, make possible cleanup and
            return False. Raise exceptions from here only if error in unrecoverable (aka fatal)

        :return: True if stopping succeed, False otherwise
        """
        # put custom termination code here, return from this method when all related objects are STOPPED
        # await if needed!
        pass

    async def recover(self) -> bool:
        """
        Recover is called by parent component if this components terminates (it states change to False)
        unexpectedly become corrupted.

        If recovering is possible try to do that here, but remember that current state can be really
        *EVERYTHING*. If recovering fail or it's not possible return False.

        Return only when recovery procedure is fully completed or failed. Awaits on internal steps.
        :return:
        """
        return False



# -*- coding: utf-8 -*-

import asyncio

import logging
import pytest
from unittest.mock import patch, AsyncMock

from patchwork.core import Component, OrphanedComponent


class Parent:
    pass


@pytest.fixture
def parent():
    return Parent()


def test_parent():

    parent = Parent()
    cmp = Component(parent=parent)
    assert parent == cmp.parent

    del parent
    with pytest.raises(OrphanedComponent):
        _ = cmp.parent


@pytest.mark.asyncio
async def test_start_procedure(parent):
    cmp = Component(parent=parent)

    with patch.object(cmp, '_start', new_callable=AsyncMock) as setup_mock:
        assert await cmp.run()

        setup_mock.assert_called()
        assert cmp.is_running

        setup_mock.reset_mock()
        assert not await cmp.run(), \
            "Starting already running component should return False"

        assert not setup_mock.called, \
            "Start method called on running component should not invoke _start()"


@pytest.mark.asyncio
async def test_stop_procedure(parent):

    cmp = Component(parent=parent)

    # inject running state
    cmp.state.set(True)

    with patch.object(cmp, '_stop', new_callable=AsyncMock) as teardown_mock:
        assert await cmp.terminate()

        teardown_mock.assert_called()
        assert not cmp.is_running

        teardown_mock.reset_mock()
        assert not await cmp.terminate(), \
            "Stopping not running component should return False"

        assert not teardown_mock.called, \
            "Stop method called on not running component should not invoke _stop()"


@pytest.mark.asyncio
async def test_state_flag_changes(parent, event_loop):

    cmp = Component(parent=parent)
    callback_got = None

    async def callback():
        nonlocal callback_got
        state = await cmp.state.wait()
        callback_got = state

    t = event_loop.create_task(callback())
    await asyncio.wait((cmp.run(), t))

    assert callback_got, \
        "Flag wait() should return True which is running component state"
    callback_got = None

    t = event_loop.create_task(callback())
    await asyncio.wait((cmp.terminate(), t))

    assert not callback_got, \
        "Flag wait() should return False which is not running component state"


@pytest.mark.asyncio
async def test_cant_run_disabled(parent, suppress_logging):

    cmp = Component(parent=parent, enabled=False)

    with pytest.raises(RuntimeError), suppress_logging(max_level=logging.ERROR):
        await cmp.run()



# -*- coding: utf-8 -*-
from ..task import Task


class BaseTaskSerializer:

    def encode(self, task: Task) -> bytes:
        raise NotImplementedError()

    def decode(self, data: bytes) -> Task:
        raise NotImplementedError()



# -*- coding: utf-8 -*-

from pydantic import BaseSettings, BaseModel, ValidationError
from pydantic.fields import ModelField
from typing import TypeVar, Generic, Type, Union, Mapping

from patchwork.core import Component, AsyncPublisher, AsyncSubscriber
from patchwork.core.config.providers import json_config_settings_source
from patchwork.core.pydantic import nested_env_settings
from patchwork.core.typing import ClassPath


class ImproperlyConfigured(Exception):
    """
    Invalid configuration detected
    """
    pass


ClassType = TypeVar("ClassType", bound=Type[object])


class ClassConfig(Generic[ClassType]):

    engine: ClassPath[ClassType]
    options: Mapping = {}

    def __init__(self, engine: ClassType, options: Mapping = None):
        self.engine = engine
        self.options = options or {}

    @classmethod
    def __get_validators__(cls):
        yield cls.dependency_validator

    @classmethod
    def dependency_validator(cls, v, field: ModelField):
        if isinstance(v, cls):
            return v

        engine_field = ModelField(name='engine', type_=ClassPath[ClassType],
                                  class_validators=None, model_config=field.model_config)
        engine, error = engine_field.validate(v.pop('engine', None), {}, loc='engine')
        if error:
            raise ValidationError([error], cls)

        # TODO: it will be nice to have validate options against constructor arguments
        opts = v
        return cls(engine=engine.type_, options=opts)

    def instantiate(self, **init_kwargs):
        return self.engine(**self.options, **init_kwargs)


ComponentType = TypeVar("ComponentType", bound=Type[Component])


class ComponentConfig(Generic[ComponentType]):
    """
    pydantic custom type for Dependencies.
    A dependency is a configuration of another component. Configuration must have at least `engine`
    defined which points to class and optionally may have a dict of `options` which will be passed
    to the engine. Options must follow engine `Settings`.
    """

    engine: ClassPath[ComponentType]
    options: BaseModel = BaseModel()

    def __init__(self, engine: ComponentType, options: Union[BaseModel, dict] = None):
        self.engine = engine
        if options is None:
            options = engine.Config()
        elif isinstance(options, dict):
            options = engine.Config(**options)
        else:
            assert isinstance(options, engine.Config), \
                f"model config for {engine} must be an instance of the component config"
        self.options = options

    @classmethod
    def __get_validators__(cls):
        yield cls.dependency_validator

    @classmethod
    def dependency_validator(cls, v, field: ModelField):
        if isinstance(v, cls):
            return v

        engine_field = ModelField(name='engine', type_=ClassPath[ComponentType],
                                  class_validators=None, model_config=field.model_config)
        engine, error = engine_field.validate(v.pop('engine', None), {}, loc='engine')
        if error:
            raise ValidationError([error], cls)

        options_field = getattr(engine.type_, 'Config', None)
        if options_field is not None:
            opts = options_field.validate(v)
        else:
            opts = {}
        return cls(engine=engine.type_, options=opts)

    def instantiate(self, parent=None, **init_kwargs):
        return self.engine(**({"parent": parent} if parent is not None else {}), **self.options.dict(), **init_kwargs)


PublisherConfig = ComponentConfig[Type[AsyncPublisher]]
SubscriberConfig = ComponentConfig[Type[AsyncSubscriber]]


class PatchworkConfig(BaseSettings):

    class Config:
        env_prefix = 'patchwork_'
        env_file = '.env'
        env_file_encoding = 'utf-8'
        json_file = 'settings.json'
        secrets_dir = None

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                json_config_settings_source,
                nested_env_settings,
                env_settings,
                file_secret_settings,
            )



# -*- coding: utf-8 -*-
import os

import pytest
from pydantic import BaseSettings, BaseModel

from patchwork.core.pydantic import nested_env_settings


class SetEnv:
    def __init__(self):
        self.envars = set()

    def set(self, name, value):
        self.envars.add(name)
        os.environ[name] = value

    def clear(self):
        for n in self.envars:
            os.environ.pop(n)


@pytest.fixture
def env():
    setenv = SetEnv()

    yield setenv

    setenv.clear()


class Settings(BaseSettings):
    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                nested_env_settings,
                file_secret_settings,
            )


def test_nested_env_delimiter(env):
    class SubSubValue(BaseModel):
        v6: str

    class SubValue(BaseModel):
        v4: str
        v5: str
        sub_sub: SubSubValue

    class TopValue(Settings):
        v1: str
        v2: str
        v3: str
        sub: SubValue

    class Cfg(Settings):
        v0: str
        top: TopValue

        class Config:
            env_nested_delimiter = '__'

    env.set('top', '{"v1": "1", "v2": "2", "sub": {"v5": "xx"}}')
    env.set('v0', '0')
    env.set('top__v3', '3')
    env.set('top__sub', '{"sub_sub": {"v6": "6"}}')
    env.set('top__sub__v4', '4')
    env.set('top__sub__v5', '5')
    cfg = Cfg()
    assert cfg.dict() == {
        'v0': '0',
        'top': {
            'v1': '1',
            'v2': '2',
            'v3': '3',
            'sub': {
                'v4': '4',
                'v5': '5',
                'sub_sub': {
                    'v6': '6'
                }
            },
        },
    }




# -*- coding: utf-8 -*-
from typing import Type

import pytest
from pydantic import BaseModel

from patchwork.core.typing import FuncPath


class Result:
    pass


def example_function():
    return Result


async def example_coroutine():
    return Result


def test_function_no_type():
    class M(BaseModel):
        path: FuncPath

    m = M(path='tests.core.test_fnpath:example_function')
    assert m.path() == Result


def test_function_default():
    class M(BaseModel):
        path: FuncPath = FuncPath(example_function)

    m = M()
    assert m.path() == Result


@pytest.mark.asyncio
async def test_coroutine_no_type():
    class M(BaseModel):
        path: FuncPath

    m = M(path='tests.core.test_fnpath:example_coroutine')
    assert await m.path() == Result


@pytest.mark.asyncio
async def test_coroutine_default():
    class M(BaseModel):
        path: FuncPath = FuncPath(example_coroutine)

    m = M()
    assert await m.path() == Result


# -*- coding: utf-8 -*-
from patchwork.core.client.serializers.base import BaseTaskSerializer
from patchwork.core.proto.google.protobuf import Value, NullValue
from patchwork.core.proto.task import ProtoTask, ProtoTaskMetadata
from patchwork.core.proto.google.protobuf import StringValue, BytesValue, Struct

try:
    import betterproto
except ImportError:
    betterproto = None

from ..task import Task, TaskMetadata


class BetterprotoSerializer(BaseTaskSerializer):

    def __init__(self):
        if betterproto is None:
            raise Exception("`betterproto` not installed")

    @classmethod
    def _encode_meta_extra(cls, v):
        if v is None:
            return Value(null_value=NullValue.NULL_VALUE)
        elif isinstance(v, (int, float)):
            return Value(number_value=v)
        elif isinstance(v, str):
            return Value(string_value=v)
        elif isinstance(v, bool):
            return Value(bool_value=v)
        else:
            raise NotImplementedError()

    @classmethod
    def _encode_payload(cls, payload):
        if isinstance(payload, str):
            wrapper = StringValue(value=payload)
            return 'google.protobuf.StringValue', bytes(wrapper)
        elif isinstance(payload, bytes):
            wrapper = BytesValue(value=payload)
            return 'google.protobuf.BytesValue', bytes(wrapper)
        else:
            try:
                return f'custom', bytes(payload)
            except TypeError:
                raise ValueError('unsupported payload type')

    @classmethod
    def _decode_payload(cls, payload):
        if payload.type_url == 'google.protobuf.StringValue':
            w = StringValue()
            w.parse(payload.value)
            return w.value
        elif payload.type_url == 'google.protobuf.BytesValue':
            w = BytesValue()
            w.parse(payload.value)
            return w.value
        elif payload.type_url == 'custom':
            # actual decoding needs to be done on client code depending on actual transfer protocol and encoding
            return payload.value
        else:
            raise ValueError(f'unsupported payload type: {payload.type_url}')

    def encode(self, task: Task) -> bytes:

        proto_task = ProtoTask(
            uuid=str(task.uuid),
            task_type=task.task_type,
            correlation_id=task.correlation_id,
        )

        proto_task.payload.type_url, proto_task.payload.value = self._encode_payload(task.payload)

        task_meta_fields = {
            'expires',
            'max_retries',
            'queue_name',
            'not_before',
            'attempt',
            'received',
            'scheduled'
        }

        raw = task.meta.dict(include=task_meta_fields)
        proto_task.meta = ProtoTaskMetadata(**raw)

        if task.meta.extra:
            extra = Struct()
            meta = {k: self._encode_meta_extra(v) for k, v in task.meta.extra.items()}
            extra.fields.update(meta)
            proto_task.meta.extra.type_url = 'google.protobuf.Struct'
            proto_task.meta.extra.value = bytes(extra)

        return bytes(proto_task)

    def decode(self, data: bytes) -> Task:
        task_proto = ProtoTask()
        task_proto.parse(data)

        proto_extra = task_proto.meta.extra
        extra = {}

        if proto_extra.type_url == 'google.protobuf.Struct':
            s = Struct()
            s.parse(proto_extra.value)
            extra = {k: betterproto.which_one_of(v, 'kind')[1] for k, v in s.fields.items()}

        task = Task(
            uuid=task_proto.uuid,
            task_type=task_proto.task_type,
            correlation_id=task_proto.correlation_id,
            payload=self._decode_payload(task_proto.payload),
            meta=TaskMetadata(
                not_before=task_proto.meta.not_before if task_proto.meta.not_before.timestamp() > 0 else None,
                expires=task_proto.meta.expires if task_proto.meta.expires.timestamp() > 0 else None,
                max_retries=task_proto.meta.max_retries,
                attempt=task_proto.meta.attempt,
                scheduled=task_proto.meta.scheduled if task_proto.meta.scheduled.timestamp() > 0 else None,
                received=task_proto.meta.received if task_proto.meta.received.timestamp() > 0 else None,
                queue_name=task_proto.meta.queue_name,
                extra=extra
            )
        )

        return task



# -*- coding: utf-8 -*-
from asyncio import AbstractEventLoop, get_event_loop

import uuid
import warnings
from datetime import datetime
from typing import Tuple, Mapping, Any, Callable, Union, Iterable, Type

import pytz

from .serializers.base import BaseTaskSerializer
from .serializers.betterproto import BetterprotoSerializer
from .task import Task, TaskMetadata, TaskDecodeError
from ..component import Component


class PatchworkPublisher(Component):
    """
    Common code for sync and async publishers
    """

    class Config(Component.Config):
        """
        Settings schema for each Patchwork Client.
        :cvar max_message_size:     Maximum allowed message size (on the wire) in bytes
        """
        max_message_size: int = 1024*1024       # 1MB
        serializer: Type[BaseTaskSerializer] = BetterprotoSerializer

    settings: Config
    asynchronous: bool
    routing: list

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.routing = []
        self.task_serializer = self.settings.serializer()

    def __repr__(self):
        return super().__repr__().replace('Component', 'Publisher')

    def add_router(self, router: Callable[[Task], Union[str, None]]):
        """
        Add a task router for the publisher. Routers will be executed in the reverse order of adding
        (the most newly added - firstly executed)
        :param router: callable which for given task returns a name of queue for it or None if this router
        does not support this task
        :return:
        """
        self.routing.insert(0, router)

    def remove_router(self, router: Callable[[Task], Union[str, None]]):
        """
        Removes previously added router
        :param router:
        :return:
        """
        self.routing.remove(router)

    def _serialize_task(self, task: Task) -> bytes:
        """
        Take task and returns serialized task payload as bytes.
        :param task:    Task instance
        :return:    Serialized task
        """
        return self.task_serializer.encode(task)

    def _prepare_task(self, task: Task) -> bytes:
        """
        Prepares task to be send. Makes validation tests on the task and returns serialized payload.
        Internally this method calls _serialize_task()
        :param task:
        :raise ValueError: if task validation fails
        :return: payload of serialized task if validation pass
        """
        if not task.uuid:
            task.uuid = str(uuid.uuid4())
        task.meta.scheduled = datetime.now(pytz.UTC)

        if task.meta.expires and task.meta.expires < task.meta.scheduled:
            raise ValueError('unable to send task which is already expired')

        if not task.meta.queue_name:
            for router in self.routing:
                queue_name = router(task)
                if queue_name is not None:
                    break

        payload = self._serialize_task(task)
        if len(payload) > self.settings.max_message_size:
            raise ValueError('message is too big')

        return payload

    def _build_task(self, payload: Any, meta: Union[Mapping, TaskMetadata], cause: Task = None, task_type: str = None):
        if not isinstance(meta, TaskMetadata):
            meta = dict(meta)
            metadata = TaskMetadata()

            if 'not_before' in meta:
                metadata.not_before = meta.pop('not_before')
            if 'expires' in meta:
                metadata.expires = meta.pop('expires')
            if 'max_retries' in meta:
                metadata.max_retries = meta.pop('max_retries')
            if 'queue_name' in meta:
                metadata.queue_name = meta.pop('queue_name')

            if meta:
                metadata.extra = dict(meta)
        else:
            metadata = meta

        task = Task(
            uuid=uuid.uuid4(),
            meta=metadata,
            task_type=task_type or payload.__class__.__name__,
            payload=payload
        )

        if cause is not None:
            task.correlation_id = cause.correlation_id or cause.uuid

        return task

    def __del__(self):
        if self.is_running:
            warnings.warn(f"Destroying unstopped publisher may leads to data loss! stop() didn't finished or called")


class PatchworkSubscriber(Component):

    class Config(Component.Config):
        """
        Settings schema for each Patchwork Client.
        """
        serializer: Type[BaseTaskSerializer] = BetterprotoSerializer

    settings: Config
    asynchronous: bool

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.task_serializer = self.settings.serializer()

    def __repr__(self):
        return super().__repr__().replace('Component', 'Subscriber')

    def _deserialize_task(self, payload: bytes, meta: Mapping) -> Task:
        """
        Deserialize given payload as a Task.
        :param payload:
        :param meta:    Additional data passed by receiver
        :return:        Deserialized task instance
        """
        return self.task_serializer.decode(payload)

    def _process_received_task(self, payload: bytes, meta: Mapping) -> Task:
        """
        Processes received payload and converts into task. Internally this method calls _deserialize_task()
        :param payload:
        :param meta:    additional informaction passed by a receiver
        :return:  deserialized task instance
        """
        try:
            task = self._deserialize_task(payload, meta)
        except Exception as exc:
            handled = self._handle_deserialize_error(payload, meta, exc)
            raise TaskDecodeError(payload, meta, exc, handled) from exc

        task.meta.received = datetime.now(pytz.UTC)
        task.meta.queue_name = meta.get('queue_name', None)
        return task

    def _handle_deserialize_error(self, payload: bytes, meta: Any, exc: Exception) -> bool:
        """
        Called when exception is raised during message deserialization and task initialization.
        This method is responsible of making sure that task won't be lost, however simple resending
        to the queue might be not accurate because if error happens once it probably will happen
        every time.
        :param payload:
        :param meta:
        :param exc:
        :return: True if message has been somehow handled and worker should not take any action. Otherwise
                 return False, so worker may report this issue using own logging mechanism
        """
        return False

    def __del__(self):
        if self.is_running:
            warnings.warn(f"Destroying unstopped subscriber may leads to data loss! stop() didn't finished or called")


class SynchronousPublisher(PatchworkPublisher):
    """
    Synchronous patchwork base client. This is a fallback client for external synchronous code
    without async support.
    """

    is_asynchronous = False

    def send(self, payload: Any, *, timeout: float = None, cause: Task = None, **meta):
        task = self._build_task(payload, meta, cause=cause)
        self.send_task(task, timeout=timeout)
        return task

    def send_task(self, task: Task, *, timeout: float = None):
        """
        Sends given task, if operation fail in given timeout raises TimeoutError.
        :param task:
        :param timeout: send operation timeout, None means no timeout
        :return:
        """
        payload = self._prepare_task(task)
        self._send(payload, task, timeout=timeout)

    def _send(self, payload: bytes, task: Task, *, timeout: float = None):
        """
        Sends given payload
        :param payload: Task payload to send
        :param timeout: Requested send operation timeout, None means no timeout
        :return:
        """
        raise NotImplementedError()


class SynchronousSubscriber(PatchworkSubscriber):

    def get(self, *, timeout: int = None):
        """
        Waits given time for next task and when arrived returns it. If no task came raises EOFError.
        :param timeout:
        :return:
        """
        payload, meta = self._fetch_one(timeout=timeout)
        return self._process_received_task(payload, meta)

    def commit(self, task, *, timeout: int = None):
        """
        Commits given task returned by get() method. If operation fail in given timeout
        raises TimeoutError.
        :param task:
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    def _fetch_one(self, *, timeout=None) -> Tuple[bytes, Mapping]:
        """
        Fetch one incoming message from the queue.
        :param timeout: Waits given time for message, if there is no message in given timeout raises TimeoutError.
                        None means no timeout, `0` means no wait.
        :return: payload of fetched message and additional information which should be passed to task deserializer
        """
        raise NotImplementedError()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.get()
        except EOFError:
            raise StopIteration()

    def subscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError()

    def unsubscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError()

    def subscription(self) -> Iterable[str]:
        raise NotImplementedError()


class AsyncPublisher(PatchworkPublisher):
    """
    Patchwork asynchronous pubslisher
    """

    is_asynchronous = True

    def __init__(self, *, parent=None, loop: AbstractEventLoop = None, **options):
        super().__init__(parent=parent, **options)

        if loop is None:
            loop = get_event_loop()

        self.loop = loop

    async def send(self, payload: Any, *, timeout: float = None, cause: Task = None, task_type: str = None, **meta):
        task = self._build_task(payload, meta, cause=cause, task_type=task_type)
        await self.send_task(task, timeout=timeout)
        return task

    async def send_task(self, task: Task, *, timeout: float = None):
        """
        Sends given task, if operation fail in given timeout raises TimeoutError.
        When this method returns it's considered as task has been delivered successfully and it's guaranteed
        that won't be lost (eg queue backend committed the message)
        :param task:
        :param timeout: send operation timeout, None means no timeout. 0 means immediatelly
        :return:
        """
        payload = self._prepare_task(task)
        await self._send(payload, task, timeout=timeout)

    async def _send(self, payload: bytes, task: Task, *, timeout: float = None):
        """
        Sends given payload

        :param payload: Task payload to send
        :param task: Task instance to send, it should be considered as **immutable**
        :param timeout: Requested send operation timeout, None means no timeout, 0 means immediately
        :return:
        """
        raise NotImplementedError()


class AsyncSubscriber(PatchworkSubscriber):

    is_asynchronous = True

    def __init__(self, *, parent=None, loop: AbstractEventLoop = None, **options):
        super().__init__(parent=parent, **options)

        if loop is None:
            loop = get_event_loop()

        self.loop = loop

    async def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self.get()
        except EOFError:
            raise StopAsyncIteration()

    async def subscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError()

    async def unsubscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError()

    def subscription(self) -> Iterable[str]:
        raise NotImplementedError()

    async def get(self, *, timeout: float = None) -> Task:
        """
        Waits given time for next task and when arrived returns it.
        :param timeout:
        :raise TimeoutError: no task came in given timeout
        :return:
        """
        payload, meta = await self._fetch_one(timeout=timeout)
        return self._process_received_task(payload, meta)

    async def commit(self, task: Task, *, timeout: float = None):
        """
        Commits given task returned by get() method.
        Meaning of "commit" depends on queue backend, but should be considered as success of task handling
        (which is not the same as success of task execution). Success of task handling means that task
        has been executed successfully or rescheduled successfully and given task can be safely removed
        from the queue. Whatever happens task won't be lost.

        :param task:
        :raise TimeoutError: commit timeout exceeded
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    async def rollback(self, task: Task, *, timeout: float = None):
        """
        Rollback given task returned by get() method.
        Meaning of "rollback" depends on queue backend, but after this operation exactly the same task should
        be delivered again. Use this with caution as exactly the same is strict, so your code will receive
        the same bytes again and again. Make sure that it doesn't lead to infinite loop.

        For retries, it's recommended to store some retries counter in task meta, send retry task explicitly
        with bumped counter and commit the failed one.
        :param task:
        :param timeout:
        :return:
        """
        raise NotImplementedError()

    async def _fetch_one(self, *, timeout: float = None) -> Tuple[bytes, Mapping]:
        """
        Fetch one incoming message from the queue.
        :param timeout: Waits given time for message, if there is no message in given timeout raises TimeoutError.
                        None means no timeout, `0` means no wait.
        :return: payload of fetched message and additional information which should be passed to task deserializer
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import json
from typing import Any


class Serializer:

    @classmethod
    def loads(cls, data: bytes) -> Any:
        raise NotImplementedError()

    @classmethod
    def dumps(cls, data: Any) -> bytes:
        raise NotImplementedError()


class NoOpSerializer(Serializer):

    @classmethod
    def loads(cls, data: Any) -> Any:
        return data

    @classmethod
    def dumps(cls, data: Any) -> Any:
        return data


class JSONSerializer(Serializer):

    @classmethod
    def loads(cls, data: bytes):
        return json.loads(data)

    @classmethod
    def dumps(cls, data: Any) -> bytes:
        return json.dumps(data).encode('utf-8')



# -*- coding: utf-8 -*-
import sys
from importlib import import_module

from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from typing import Union, Text, Any, TypeVar, Generic, Type

# standard python types serializable by all libraries without and special processing
from pydantic.validators import dict_validator

StandardTypes = Union[None, bytes, str, tuple, list, dict, int, float, bool]


T = TypeVar('T')


class ClassPath(Generic[T]):
    """
    String which represents a path to class, contains module path and class name delimited by a colon.
    In Pydantic returns imported class

    !!! example
        `foo.bar:MyClass`
    """

    def __init__(self, cls: Type[T]):
        self._class = cls

    @property
    def type_(self) -> Type[T]:
        return self._class

    @classmethod
    def __get_validators__(cls):
        yield cls.import_class

    @classmethod
    def import_class(cls, v, field: ModelField):
        if isinstance(v, cls):
            return v

        if v.count(':') != 1:
            raise ValueError(f"'{v}' is not a valid path to class. Path should be a Python "
                             f"module and class name joined by colon, eg: foo.bar:MyClass")

        mod_name, class_name = v.split(':')
        try:
            mod = import_module(mod_name)
        except Exception as e:
            exc = ImportError(f'unable to import {mod_name}: {e.__class__.__name__}({e})')
            raise ValidationError([
                ErrorWrapper(exc=exc, loc='cls')
            ], cls)

        for module_name, module in sys.modules.items():
            # skip empty modules and ones without actual file
            if not module or not hasattr(module, '__file__'):
                continue

            if mod.__file__ == module.__file__:
                # if same module has been imported use imported one
                # this allows to use issubclass or isinstance checks on previously imported module and
                # class returned by this method (note: same class code imported in different way creates
                # different class types)
                if mod != module:
                    sys.modules.pop(mod.__name__)
                    mod = module
                break

        if not hasattr(mod, class_name):
            raise ValueError(f"module '{mod_name}' has no member '{class_name}'")

        imported_cls = getattr(mod, class_name)

        if field.sub_fields is not None:
            cls_f = field.sub_fields[0]
            imported_cls, error = cls_f.validate(imported_cls, {}, loc='cls')
            if error:
                raise ValidationError([error], cls)

        return cls(imported_cls)

    def __call__(self, *args, **kwargs) -> T:
        return self._class(*args, **kwargs)

    def __repr__(self):
        return f"ClassPath[{self._class.__module__}:{self._class.__name__}]"


class FuncPath:
    """
    String which represents a path to function, contains module path and class name delimited by a colon.
    In Pydantic returns imported function

    !!! example
        `foo.bar:my_func`
    """

    def __init__(self, fn):
        self._fn = fn

    @classmethod
    def __get_validators__(cls):
        yield cls.import_function

    @classmethod
    def import_function(cls, v, field: ModelField):
        if not isinstance(v, str):
            return v

        if v.count(':') != 1:
            raise ValueError(f"'{v}' is not a valid path to function. Path should be a Python "
                             f"module and member name joined by colon, eg: foo.bar:my_fn")

        mod_name, member_name = v.split(':')
        mod = import_module(mod_name)

        for module_name, module in sys.modules.items():
            # skip empty modules and ones without actual file
            if not module or not hasattr(module, '__file__'):
                continue

            if mod.__file__ == module.__file__:
                # if same module has been imported use imported one
                # this allows to use issubclass or isinstance checks on previously imported module and
                # class returned by this method (note: same class code imported in different way creates
                # different class types)
                if mod != module:
                    sys.modules.pop(mod.__name__)
                    mod = module
                break

        if not hasattr(mod, member_name):
            raise ValueError(f"module '{mod_name}' has no member '{member_name}'")

        member = getattr(mod, member_name)
        return cls(member)

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)

    def __repr__(self):
        return f"FuncPath[{self._fn.__module__}:{self._fn.__name__}]"


class LoggingConfig(dict):
    """
    Python `logging` module configuration.
    """

    __origin__ = dict
    __args__ = [Text, Any]  # type: ignore

    @classmethod
    def __get_validators__(cls):
        yield dict_validator
        yield cls.logging_config_validator

    @classmethod
    def logging_config_validator(cls, v):
        # TODO:
        return v



from os.path import join, abspath, dirname

import subprocess

if __name__ == '__main__':
    root = abspath(join(dirname(__file__), '..'))
    src_dir = join(root, 'protocol')
    dst_dir = join(root, 'patchwork/core/proto')
    subprocess.run(["protoc", f"-I={src_dir}", f"--python_betterproto_out={dst_dir}", join(src_dir, "task.proto")])



# -*- coding: utf-8 -*-

from pathlib import Path

import json

from pydantic import BaseSettings
from typing import Dict, Any


def json_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    pydantic settings provider for JSON files
    :param settings:
    :return:
    """
    if settings.__config__.json_file is None:
        return {}
    path = Path(settings.__config__.json_file)
    if not path.exists():
        return {}
    return json.loads(path.read_text('utf-8'))



# -*- coding: utf-8 -*-

import asyncio
import collections


class cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    A cached property can be made out of an existing method:
    (e.g. ``url = cached_property(get_absolute_url)``).
    """
    name = None

    @staticmethod
    def func(instance):
        raise TypeError(
            'Cannot use cached_property instance without calling '
            '__set_name__() on it.'
        )

    def __init__(self, func, name=None):

        func_name = func.__name__
        name = name or func_name
        if not (isinstance(name, str) and name.isidentifier()):
            raise ValueError(
                "%r can't be used as the name of a cached_property." % name,
            )

        self.name = name
        self.func = func
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        if name != self.name:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                "(%r and %r)." % (self.name, name)
            )

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


class Flag:
    """
    Simple flag implementation, works like Event but allows to pass a value.
    Use wait() to wait for value change.
    """

    def __init__(self, initial_value=None, *, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._waiters = collections.deque()
        self._value = initial_value

    async def wait(self):
        """
        Waits for flag value change and returns new one.
        :return: flag value
        """
        f = self._loop.create_future()
        self._waiters.append(f)
        return await f

    async def wait_for(self, value):
        """
        Wait for given flag value
        :param value:
        :return:
        """
        if self._value == value:
            return

        while True:
            f = self._loop.create_future()
            self._waiters.append(f)
            res = await f
            if res == value:
                return

    def set(self, value):
        """
        Sets new flag value and notifies all waiters.
        :param value:
        :return:
        """
        if value == self._value:
            # setting to the same value is no-op
            return

        self._value = value

        # store current waiters and create new internal queue before resolving futures,
        # so if someone waits for certain value may safely re-append to waiters list
        waiters = self._waiters
        self._waiters = collections.deque()

        for f in waiters:
            if not f.cancelled():
                f.set_result(value)

    @property
    def value(self):
        """
        Gets current flag value
        :return:
        """
        return self._value

    def __repr__(self):
        res = super().__repr__()
        extra = repr(self._value)
        if self._waiters:
            extra = f'{extra}, waiters:{len(self._waiters)}'
        return f'<{res[1:-1]} [{extra}]>'


class AsyncQueue:
    """
    Async queue which works on events to block and await for empty slots or available data.
    These events are exposed so external code may check if there is a possibility to put or
    get data and await on their own.

    This is FIFO queue.

    Custom events exposed:
    - empty: set when queue becomes empty
    - nonempty: negation of empty, set when queue is not empty and has any data
    - full: set when queue becomes full and there is no free slots available
    - nonfull: negation of full, set when full is not empty and has free slots to put more data

    Interface is compatible with standard Python queues.

    """

    def __init__(self, maxsize=None):
        self.maxsize = maxsize

        self._items = collections.deque()
        self._putters = collections.deque()

        self.empty = asyncio.Event()
        self.nonempty = asyncio.Event()

        self.nonfull = asyncio.Event()
        self.full = asyncio.Event()

        self.empty.set()
        self.nonfull.set()

    def get_nowait(self):
        """
        Pops and returns item from the queue. If there is no item to return raises QueueEmpty exception
        like standard asyncio queue.
        :return:
        """
        if not self._items:
            raise asyncio.QueueEmpty()

        item = self._items.popleft()
        self.full.clear()
        self.nonfull.set()

        if not self._items:
            self.empty.set()
            self.nonempty.clear()

        return item

    def task_done(self):
        """
        Marks fetched task done.
        :return:
        """
        pass

    async def get(self, timeout: float = None):
        """
        Pops and returns item from the queue. If there is no items to return awaits for it.
        :return:
        """
        await asyncio.wait_for(self.nonempty.wait(), timeout=timeout)
        return self.get_nowait()

    async def put(self, item, timeout: float = None):
        """
        Put an item to the queue. If there are no free slots available awaits for it.
        :param item:
        :param timeout:
        :return:
        """
        await asyncio.wait_for(self.nonfull.wait(), timeout=timeout)
        return self.put_nowait(item)

    def put_nowait(self, item):
        """
        Put an item to the queue. If there is no free slots available raises QueueFull exception like
        standard asyncio queue.
        :param item:
        :return:
        """
        if len(self._items) == self.maxsize:
            raise asyncio.QueueFull()

        self._items.append(item)
        self.empty.clear()
        self.nonempty.set()

        if len(self._items) == self.maxsize:
            self.full.set()
            self.nonfull.clear()

    def revoke_nowait(self, item):
        self._items.remove(item)
        self.full.clear()
        self.nonfull.set()

        if not self._items:
            self.empty.set()
            self.nonempty.clear()

    def clear(self):
        """
        Clears the queue
        :return:
        """
        self._items.clear()
        self.full.clear()
        self.nonfull.set()
        self.nonempty.clear()
        self.empty.set()



# -*- coding: utf-8 -*-
from typing import Type
from pydantic import BaseModel

from patchwork.core.typing import ClassPath


class Test:
    pass


def test_without_type():
    class M(BaseModel):
        path: ClassPath

    m = M(path='tests.core.test_classpath:Test')


def test_with_type():
    class M(BaseModel):
        path: ClassPath[Type[Test]]

    m = M(path='tests.core.test_classpath:Test')


def test_default():
    class M(BaseModel):
        path: ClassPath[Type[Test]] = ClassPath(Test)

    m = M()
    assert issubclass(m.path.type_, Test)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: task.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime

import betterproto

from .google import protobuf


@dataclass
class ProtoTaskMetadata(betterproto.Message):
    not_before: datetime = betterproto.message_field(1)
    expires: datetime = betterproto.message_field(2)
    max_retries: int = betterproto.uint32_field(3)
    attempt: int = betterproto.uint32_field(4)
    scheduled: datetime = betterproto.message_field(5)
    received: datetime = betterproto.message_field(6)
    queue_name: str = betterproto.string_field(7)
    extra: protobuf.Any = betterproto.message_field(10)


@dataclass
class ProtoTask(betterproto.Message):
    uuid: str = betterproto.string_field(1)
    meta: "ProtoTaskMetadata" = betterproto.message_field(2)
    task_type: str = betterproto.string_field(3)
    correlation_id: str = betterproto.string_field(4)
    payload: protobuf.Any = betterproto.message_field(10)



# -*- coding: utf-8 -*-
import uuid as uuid
from datetime import datetime

from pydantic import BaseModel, UUID4, Field, Extra, PrivateAttr
from typing import Mapping, Optional, Any, Union, Dict


class TaskDecodeError(ValueError):
    """
    Received message cannot be decoded into task.
    """

    def __init__(self, payload: bytes, meta: Mapping, exc: Exception, handled: bool):
        """
        :param payload: Received payload
        :param meta:    Received meta
        :param exc:     Raised exception
        :param handled: True if exception has been handled by client and this is just a notification exception,
                        handled means that task is not lost on the queue. If False, this case is not handled
                        and worker should take care of logging and administrator notification in the same
                        way as other task processing messages are handled.
        """
        self.payload = payload
        self.meta = meta
        self.exception = exc
        self.is_handled = handled

    def __str__(self):
        return ("Handled" if self.is_handled else "Unhandled") + f" {self.__class__.__name__}({self.exception})"


class TaskMetadata(BaseModel):
    not_before: Optional[datetime]
    expires: Optional[datetime]
    max_retries: Optional[int]
    attempt: int = 0
    scheduled: Optional[datetime]
    received: Optional[datetime]
    queue_name: Optional[str]
    extra: Optional[Mapping[str, Union[int, float, bool, str, None]]]


class Task(BaseModel):
    uuid: UUID4 = Field(default_factory=uuid.uuid4)
    meta: TaskMetadata = TaskMetadata()
    task_type: str
    correlation_id: Optional[str]
    payload: Any

    # this field SHOULD NOT be transferred
    _local: Dict = PrivateAttr(default_factory=dict)



# -*- coding: utf-8 -*-

import asyncio
import warnings
from collections import defaultdict, deque
from enum import IntEnum
from typing import List, Dict, Any, AsyncGenerator, Iterable, TypeVar, Generic, Tuple, Mapping

from .base import SynchronousPublisher, SynchronousSubscriber, AsyncPublisher, AsyncSubscriber


class SyncLocalPublisher(SynchronousPublisher):
    """
    Patchwork is asynchronous framework with fully async nodes, so locally async Python must be supported
    to run the worker. Synchronous client make no sense in that case.
    """
    def __init__(self):
        raise NotImplementedError("There is no synchronous local client")


class SyncLocalSubscriber(SynchronousSubscriber):
    """
    Patchwork is asynchronous framework with fully async nodes, so locally async Python must be supported
    to run the worker. Synchronous client make no sense in that case.
    """
    def __init__(self):
        raise NotImplementedError("There is no synchronous local client")


class DummySerializer:

    @classmethod
    def dumps(cls, data):
        return data

    @classmethod
    def loads(cls, data):
        return data


class MissingBehaviour(IntEnum):
    # skip missing queue
    SKIP = 0
    # warn that queue is missing and skip
    WARN = 1
    # create missing queue
    CREATE = 2
    # raise with ValueError exception
    EXCEPTION = 3
    # wait until queue will be created by another code
    WAIT = 4


class DROP:
    pass


T = TypeVar('T')


class Topic(Generic[T]):
    """
    A queue implementation similar to asyncio one, but works with multiple asyncio loops as queue itself
    is not bound to any loop
    """

    def __init__(self, maxsize: int = None):
        self._queue = deque(maxlen=maxsize)
        self._get_waiters = deque()
        self._put_waiters = deque()
        self._unfinished_tasks = 0

    def qsize(self):
        return len(self._queue)

    def get_nowait(self) -> T:
        item = self._queue.popleft()
        self._unfinished_tasks += 1
        if self._put_waiters:
            self._put_waiters.popleft().set()

        return item

    def put_nowait(self, item: T):
        self._queue.append(item)
        if self._get_waiters:
            self._get_waiters.popleft().set()

    async def get(self) -> T:
        if self._queue:
            return self.get_nowait()

        event = asyncio.Event()
        self._get_waiters.append(event)
        try:
            await event.wait()
        except asyncio.CancelledError:
            self._get_waiters.remove(event)
            raise

        return await self.get()

    async def put(self, item: T):
        if len(self._queue) < self._queue.maxlen:
            return self.put_nowait(item)

        event = asyncio.Event()
        self._put_waiters.append(event)
        try:
            await event.wait()
        except asyncio.CancelledError:
            self._put_waiters.remove(event)
            raise

        return await self.put(item)

    def task_done(self):
        if self._unfinished_tasks == 0:
            raise ValueError()
        self._unfinished_tasks -= 1


class AsyncLocalBroker:
    """
    Simple local broker working on asyncio loop for testing and development purposes only.

    !!! danger
        For development purposes only!
    """

    __default = []

    def __init__(
            self,
            initial_queues: Iterable[str] = tuple(),
            publish_missing: MissingBehaviour = MissingBehaviour.CREATE,
            subscribe_missing: MissingBehaviour = MissingBehaviour.WAIT,
            max_queue_size: int = 100
    ):
        self._queues: Dict[str, Topic] = {}
        self._publish_missing = publish_missing
        self._subscribe_missing = subscribe_missing
        self._queue_size = max_queue_size

        self._missing_notif: Dict[str, List[asyncio.Event]] = defaultdict(list)

        for q_name in initial_queues:
            self._create_queue(q_name)

    @classmethod
    def default(cls):
        if not cls.__default:
            cls.__default.append(cls())

        return cls.__default[0]

    def _create_queue(self, name):
        self._queues[name] = Topic(maxsize=self._queue_size)
        if name in self._missing_notif:
            for n in self._missing_notif.pop(name):
                n.set()

    def get_queue(self, name: str) -> Topic:
        return self._queues[name]

    def has_queue(self, name: str) -> bool:
        return name in self._queues

    async def put(self, msg: Any, *, queue_name: str):
        if queue_name not in self._queues:
            if self._publish_missing == MissingBehaviour.SKIP:
                return
            elif self._publish_missing == MissingBehaviour.WARN:
                warnings.warn(
                    f'{self.__class__.__name__}: unable to deliver message to queue {queue_name}: no such queue'
                )
                return
            elif self._publish_missing == MissingBehaviour.CREATE:
                self._create_queue(queue_name)
            elif self._publish_missing == MissingBehaviour.WAIT:
                await self._missing_created(queue_name)
            else:
                raise ValueError(f'{queue_name}: no such queue')

        await self._queues[queue_name].put(msg)

    def _missing_created(self, queue_name):
        event = asyncio.Event()
        self._missing_notif[queue_name].append(event)

        async def notifier():
            await event.wait()
            return DROP
        return notifier()

    async def subscribe(self, queue_names: List[str]) -> AsyncGenerator:
        if not queue_names:
            return

        waiters: Dict[asyncio.Task, str] = {}

        for queue_name in queue_names:
            if queue_name not in self._queues:
                if self._subscribe_missing == MissingBehaviour.SKIP:
                    return
                elif self._subscribe_missing == MissingBehaviour.WARN:
                    warnings.warn(
                        f'{self.__class__.__name__}: unable to subscribe on queue {queue_name}: no such queue')
                    return
                elif self._subscribe_missing == MissingBehaviour.CREATE:
                    self._create_queue(queue_name)
                elif self._subscribe_missing == MissingBehaviour.WAIT:
                    waiters[asyncio.create_task(self._missing_created(queue_name))] = queue_name
                else:
                    raise ValueError(f'{queue_name}: no such queue')
            else:
                waiters[asyncio.create_task(self._queues[queue_name].get())] = queue_name

        while True:
            try:
                done, pending = await asyncio.wait(waiters.keys(), return_when=asyncio.FIRST_COMPLETED)

                for fut in done:
                    # remove resolved future from dict, get associated queue name
                    q_name = waiters.pop(fut)
                    # create new waiter task for this queue name
                    waiters[asyncio.create_task(self._queues[q_name].get())] = q_name
                    # yield fetched value
                    result = fut.result()
                    if result is not DROP:
                        yield result, q_name
                        self._queues[q_name].task_done()
            except asyncio.CancelledError:
                for w in waiters.keys():
                    if not w.done():
                        w.cancel()
                raise
            except GeneratorExit:
                for w in waiters.keys():
                    w.cancel()

                return

    def __str__(self):
        return f'<{self.__class__.__name__}: {", ".join(self._queues.keys())}>'


class AsyncLocalPublisher(AsyncPublisher):
    """
    Simple patchwork client working on local event loop using given local broker

    !!! danger
        For development purposes only!
    """

    def __init__(self, parent=None, broker: AsyncLocalBroker = None, **options):
        """
        :param queue:   asyncio queue to bind to
        """
        super().__init__(parent=parent, **options)
        if broker is None:
            broker = AsyncLocalBroker.default()

        self._broker = broker

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-1]}, broker={self._broker}]>"

    @property
    def broker(self):
        return self._broker

    async def _start(self):
        self.logger.debug(f"Publisher attached to broker {self._broker}")

    async def _stop(self):
        self.logger.debug(f"Publisher left broker {self._broker}")

    async def _send(self, payload, task, timeout: float = None):
        assert task.meta.queue_name, "missing task queue name"
        try:
            if timeout == 0:
                await self._broker.put(payload, queue_name=task.meta.queue_name)
            else:
                await asyncio.wait_for(self._broker.put(payload, queue_name=task.meta.queue_name), timeout=timeout)
        except asyncio.TimeoutError:
            raise TimeoutError(f"send operation timeout, can't deliver in {timeout}s")


class AsyncLocalSubscriber(AsyncSubscriber):

    class Config(AsyncSubscriber.Config):
        queue_names: List[str] = []

    def __init__(self, parent=None, broker: AsyncLocalBroker = None, **options):
        """
        :param queue:   asyncio queue to bind to
        """
        if broker is None:
            broker = AsyncLocalBroker.default()

        self._broker = broker
        super().__init__(parent=parent, **options)
        self.uncommitted = set()
        self._fetcher: AsyncGenerator

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-1]}, broker={self._broker}]>"

    @property
    def broker(self):
        return self._broker

    async def _start(self):
        self.logger.debug(f"Subscriber attached to broker {self._broker}")
        if not self.settings.queue_names:
            warnings.warn(f'{self.__class__.__name__}: no queues to listen on')
        self._fetcher = self._broker.subscribe(self.settings.queue_names)

    async def _stop(self):
        await self._fetcher.aclose()
        self.logger.debug(f"Subscriber left broker {self._broker}")

    async def _fetch_one(self, timeout: float = None) -> Tuple[bytes, Mapping]:
        try:
            data, q_name = await asyncio.wait_for(self._fetcher.__anext__(), timeout=timeout)
            return data, {'queue_name': q_name}
        except AttributeError:
            if not hasattr(self, '_fetcher'):
                raise RuntimeError(
                    "Can't fetch task: subscriber seems to be not started. "
                    "Did you forgot to call run() method and await?"
                )
            raise
        except asyncio.TimeoutError:
            raise TimeoutError(f"fetch operation timeout, no messages in {timeout}s")

    async def commit(self, task, *, timeout: float = None):
        self.uncommitted.remove(task.uuid)

    async def rollback(self, task, *, timeout: float = None):
        pass

    async def get(self, *, timeout: float = None):
        task = await super().get(timeout=timeout)
        self.uncommitted.add(task.uuid)
        return task

    async def subscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError('changing subscription on local client is not supported')

    async def unsubscribe(self, queue_name: Iterable[str]):
        raise NotImplementedError('changing subscription on local client is not supported')

    def subscription(self) -> Iterable[str]:
        return self.settings.queue_names
